@@ -1,5 +1,5 @@
 ;; scss.scm: main module exports and implementations for SCSS
-;; Copyright (C) 2008 Julian Graham
+;; Copyright (C) 2009 Julian Graham
 
 ;; SCSS is free software: you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
@@ -41,7 +41,10 @@
 	    scss:parser-debug-messages
 
 	    scss:css->scss
-	    scss:scss->css)
+	    scss:scss->css
+	    
+	    font-size?
+	    line-height?)
 
   #:use-module (srfi srfi-1)
   #:use-module (srfi srfi-2)
@@ -49,10 +52,26 @@
   #:use-module (srfi srfi-13)
   #:use-module (srfi srfi-14)
   #:use-module (srfi srfi-69))
- 
-(load "lexer.scm")
-(load "parser.scm")
-(load "srfi-32.scm")
+
+(primitive-load-path "scss/lexer")
+(primitive-load-path "scss/parser")
+(primitive-load-path "scss/srfi-32")
+
+(define-record-type lexical-token
+  (make-lexical-token category source value)
+  lexical-token?
+  (category lexical-token-category)
+  (source   lexical-token-source)
+  (value    lexical-token-value))
+
+(define-record-type source-location
+  (make-source-location input line column offset length)
+  source-location?
+  (input   source-location-input)
+  (line    source-location-line)
+  (column  source-location-column)
+  (offset  source-location-offset)
+  (length  source-location-length))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;
@@ -82,10 +101,9 @@
 
 (define (length? x) 
   (define (f y) (string-suffix? y x))
-  (and (find f '("em" "ex" "px" "in" "cm" "mm" "pt" "pc"))
-       (css-number? (substring x 0 (- (string-length x) 2)))))
-
-(define (pos-length? x) (or (equal? x "0") (length? x)))
+  (or (and (find f '("em" "ex" "px" "in" "cm" "mm" "pt" "pc"))
+	   (css-number? (substring x 0 (- (string-length x) 2))))
+      (equal? x "0")))
 
 (define (cue? x) 
   (every? (lambda (x) (or (uri? x) (member x '("none" "inherit"))))
@@ -183,7 +201,7 @@
 			     (member x '("normal" "inherit"))))
 
 ;; This isn't 100% accurate, but...
-(define (uri? x) (string-prefix? "(url" x))
+(define (uri? x) (string-prefix? "url(" x))
 (define (ls-image? x) (or (uri? x) (equal? x "none")))
 (define ls-positions '("inside" "outside"))
 (define ls-types '("disc" "circle" "square" "decimal" "decimal-leading-zero" 
@@ -193,7 +211,11 @@
 
 (define (margin? x) (or (length? x) (percentage? x) (equal? x "auto")))
 
-(define (counter? x) (string-prefix? "(counter" x))
+(define (counter? x) 
+  (and (string-prefix? "counter(" x)
+       (let* ((args (substring x 8 (- (string-length x) 1)))
+	      (sx (string-tokenize args (char-set #\,))))
+	 (every ident? sx))))
 (define (ocolor? x) (or (color? x) (equal? x "invert")))
 
 (define (css-time? x) (or (string-suffix? "ms" x)
@@ -263,15 +285,24 @@
 (define (clip? x) 
   (or (member x '("auto" "inherit")) (string-prefix? "(rect" x)))
 (define (content? x)
-  (define (f x)
-    (or (css-string? x)
-	(uri? x)
-	(counter? x)
-	(member x '("open-quote" 
-		    "close-quote" 
-		    "no-open-quote" 
-		    "no-close-quote"))))
-  (or (member x '("normal" "inherit")) (every f (string-tokenize x))))
+  (define (f x str-chr)
+    (or (null? x)
+	(let ((cx (car x)))
+	  (if str-chr
+	      (if (eqv? (string-ref cx (- (string-length cx) 1)) str-chr)
+		  (f (cdr x) #f)
+		  (f (cdr x) str-chr))
+	      (let ((sc (string-ref cx 0)))
+		(if (memv sc '(#\" #\'))
+		    (f (cdr x) sc)
+		    (and (or (uri? cx)
+			     (counter? cx)
+			     (member cx '("open-quote"
+					  "close-quote"
+					  "no-open-quote"
+					  "no-close-quote")))
+			 (f (cdr x) #f))))))))
+  (or (member x '("normal" "inherit")) (f (string-tokenize x) #f)))
 (define (counter-*? x) 
   (or (ident? x) (css-integer? x) (member x '("none" "inherit"))))
 (define (cursor? x)
@@ -370,6 +401,13 @@
   (or (css-number? x) 
       (member x '("x-slow" "slow" "medium" "fast" "x-fast" "faster" "slower"))
       (equal? x "inherit")))
+(define (system-font? x)
+  (member x '("caption" 
+	      "icon" 
+	      "menu" 
+	      "message-box" 
+	      "small-caption" 
+	      "status-bar")))
 (define (text-align? x) 
   (member x '("left" "right" "center" "justify" "inherit")))
 (define (text-decoration? x)
@@ -475,28 +513,26 @@
     ((font-style) ,font-style? ("normal") #t)
     ((font-variant) ,font-variant? ("normal") #t)
     ((font-weight) ,font-weight? ("normal") #t)
-    ((font) ,(lambda (x) 
+    ((font) ,(lambda (x)
 	       (let ((f (lambda (x) (or (font-style? x)
 					(font-variant? x)
 					(font-weight? x))))
-		     (g (lambda (x) (font-size? x))))
-		 (or (member x '("caption"
-				 "icon"
-				 "menu"
-				 "message-box"
-				 "small-caption"
-				 "status-bar"
-				 "inherit"))
-		     (let* ((sx (string-tokenize x))
-			    (lc (last sx)))
-		       (and (font-family? lc)
-			    (let ((sx (delete lc sx)))
-			      (case (length sx)
-				((0) #t)
-				((1) (let ((c (car sx))) (or (f c) (g c))))
-				((2) (and (f (car sx)) (g (cadr sx)))))))))))
-	  ("normal" "normal" "normal" "medium" "monospace") 
-	  #t)
+		     (g (lambda (x)
+			  (let ((si (string-index x #\/)))
+			    (if si 
+				(let ((a (substring x 0 si))
+				      (b (substring x (+ si 1))))
+				  (and (font-size? a)
+				       (line-height? b)))
+				(font-size? x))))))
+		 (or (system-font? x)
+		     (equal? x "inherit")		     
+		     (let ((sx (string-tokenize x)))
+		       (and (>= (length sx) 2)
+			    (g (car sx))
+			    (font-family? (cadr sx)))))))
+          ("normal" "normal" "normal" "medium" "monospace") 
+   	  #t)
     ((height left right width bottom top) ,height? ("auto") #f)
     ((letter-spacing) ,letter-spacing? ("normal") #t)
     ((line-height) ,line-height? ("normal") #t)
@@ -674,7 +710,7 @@
 	 (p (or (and p (or (not quot) (< p quot)) p)
 		(list-index equal-pred chrs))))
     (if p
-	(list (case (list-ref chrs p) ((#\~) '~=) ((#\|) '\|=) ((#\=) '=))
+	(list (case (list-ref chrs p) ((#\~) '~=) ((#\|) '|=) ((#\=) '=))
 	      (string->symbol (list->string (take (cdr chrs) (- p 1)))) 
 	      (let* ((ccs (cdr chrs))
 		     (val (list-tail ccs (if (eqv? (list-ref chrs p) #\=) 
@@ -685,21 +721,36 @@
 	(string->symbol 
 	 (list->string (take (cdr chrs) (- (length chrs) 2)))))))
 
-(define (add-modifiers sel modifiers)
+(define (add-specifiers sel specifiers)
   (define (contains-special-chars? str)
-    (let* ((chars (string->list str))
-	   (cchars (if (null? chars) (list) (cdr chars))))
-      (or (memv #\. cchars)
-	  (memv #\: cchars)
-	  (memv #\" cchars)
-	  (memv #\' cchars)
-	  (and (pair? cchars)
-	       (not (eqv? (car chars) #\|))
-	       (or (memv #\| cchars) (memv #\# cchars))))))
+    (define special-charset (list->char-set '(#\. #\: #\| #\[ #\' #\")))
+    (define (contains-unescaped-special-chars? str start end)
+      (let ((slash (string-index str #\\ start end)))
+	(if (and slash (<= (+ slash 2) end))
+	    (or (contains-unescaped-special-chars? str start slash)
+		(contains-unescaped-special-chars? str (+ slash 2) end))
+	    (string-index str special-charset start end))))
+    (define (contains-special-chars-pivot? start x len end)
+      (or (string-index str special-charset start x)
+	  (string-index str special-charset (+ x len) end)))
+    (let ((len (string-length str)))
+      (case (string-ref str 0)
+	((#\.) (string-index str special-charset 1 len))
+	((#\:) (string-index str special-charset 1 len))
+	((#\|) (string-index str special-charset 2 (- len 1)))
+	((#\[) 
+	 (cond ((string-contains str "|=") => 
+		(lambda (x) (contains-special-chars-pivot? 1 x 2 (- len 1))))
+	       ((string-contains str "~=") =>
+		(lambda (x) (contains-special-chars-pivot? 1 x 2 (- len 1))))
+	       ((string-contains str "=") =>
+		(lambda (x) (contains-special-chars-pivot? 1 x 1 (- len 1))))
+	       (else (string-index str special-charset 1 len))))
+	(else (string-index str special-charset 0 len)))))
 
-  (define (add-modifier modifier s)
-    (let ((sm (substring modifier 1)))
-      (case (string-ref modifier 0)
+  (define (add-specifier specifier s)
+    (let ((sm (substring specifier 1)))
+      (case (string-ref specifier 0)
 	((#\.) (list 'class s (string->symbol sm)))
 	((#\:) (let ((x (string->symbol sm)))
 		 (cond ((member x pseudo-classes) (list 'pclass s x))
@@ -707,24 +758,25 @@
 		       (else s))))
 	((#\|) `(id ,s ,(string->symbol 
 			 (substring sm 1 (string-index sm #\|)))))
-	((#\[) `(attrib ,s ,(parse-attrib-str modifier)))
+	((#\[) `(attrib ,s ,(parse-attrib-str specifier)))
 	(else s))))
 
-  (define (valid-modifier? m)
+  (define (valid-specifier? m)
     (if (eqv? (string-ref m 0) #\:)
 	(let ((n (substring m 1)))
 	  (or (member n pseudo-classes) (member n pseudo-elements)))
 	#t))
 
-  (let ((mods (if (list? modifiers) 
-		  (map replace-escapes modifiers)
-		  (map replace-escapes (list modifiers)))))
-    (if (every valid-modifier? mods)
-	(if (and (not (contains-special-chars? sel)) 
-		 (not (find contains-special-chars? mods)))
-	    (string->symbol (string-append sel (apply string-append mods)))
-	    (fold add-modifier (string->symbol sel) mods))
-	(list))))
+  (let ((mods (if (list? specifiers) 
+		  (map replace-escapes specifiers)
+		  (map replace-escapes (list specifiers))))
+	(sel (and sel (replace-escapes sel))))
+    (and (every valid-specifier? mods)
+	 (if (and (not (and sel (contains-special-chars? sel)))
+		  (not (find contains-special-chars? mods)))
+	     (string->symbol (string-append (or sel "") 
+					    (apply string-append mods)))
+	     (fold add-specifier (string->symbol (or sel "*")) mods)))))
 
 (define pseudo-classes
   '("first-child" "link" "visited" "hover" "active" "focus" "lang"))
@@ -774,11 +826,11 @@
 			((null? cs) (append `(,c) (f (cdr l) #f #f '())))
 			(else (append (list->char cs)
 				      `(,c) (f (cdr l) #f #f '()))))))))
-  (let ((r (if (symbol? s)
-	       (string->symbol (list->string 
-				(f (string->list (symbol->string s)) #f #f '())))
-	       (list->string (f (string->list s) #f #f '())))))
-    r))
+  (if (symbol? s)
+      (string->symbol 
+       (list->string 
+	(f (string->list (symbol->string s)) #f #f '())))
+      (list->string (f (string->list s) #f #f '()))))
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;
@@ -1035,7 +1087,8 @@
   (let ((pe (hash-table-ref/default property-hash-table 
 				    (symbol-downcase property)
 				    #f)))
-    (and pe (apply (css-property:match-fn pe) `(,(string-downcase value))))))
+    (and pe (apply (css-property:match-fn pe) 
+		   `(,(string-downcase value))))))
 
 (define (verify-arg-types name predicates args startnum)
   (if (not (= (length predicates) (length args)))
@@ -1049,85 +1102,20 @@
 				   (number->string (+ startnum 1)))))))))
     (f predicates args 0)))
 
-(define scss:parser-debug-messages #f)
+(define scss:parser-debug-messages #t)
 (define (display-list . lst) (if scss:parser-debug-messages (display lst)))
-
-(define (make-parser port)
-  (let* 
-      ((error-state #f)
-       (parse-error-handler (lambda args 
-			      (display-list args) 
-			      (set! error-state #t)))
-       (lexer (make-css21-lexer port))
-       (pre-token #f)
-       (use-pre-token #f)
-       (post-tokens (list))
-       (push (lambda (sym) (set! post-tokens (cons sym post-tokens))))
-       (pop (lambda (sym) (if (eq? (car post-tokens) sym)
-			      (set! post-tokens (cdr post-tokens)))))
-       (lexer-wrapper
-	(lambda ()
-	  (if use-pre-token
-	      (begin (set! use-pre-token #f) pre-token)
-	      (let* ((tok (lexer))
-		     (tok
-		      (case (car tok)
-			((LBRACE) (push 'RBRACE) tok)
-			((LPAREN) (push 'RPAREN) tok)
-			((FUNCTION) (push 'RPAREN) tok)
-			((LBRACK) (push 'RBRACK) tok)		    
-			((RBRACE) (pop 'RBRACE) tok)
-			((RPAREN) (pop 'RPAREN) tok)
-			((RBRACK) (pop 'RBRACK) tok)
-			((BAD_AT_KEYWORD)
-			 (letrec 
-			     ((f (lambda (wfb)
-				   (let ((t (lexer)))
-				     (cond ((eq? (car t) '*eoi*) t)
-					   ((and (eq? (car t) 'RBRACE))
-					    (if (<= wfb 1) 
-						(lexer) 
-						(f (- wfb 1))))
-					   ((and (eqv? wfb 0)
-						 (eq? (car t) 'SEMICOLON))
-					    (lexer))
-					   ((eq? (car t) 'LBRACE)
-					    (f (+ wfb 1)))
-					   (else (f wfb)))))))
-			   (f 0)))
-			(else tok))))
-		(let ((ctok (car tok)))
-		  (if (and ctok (not (eqv? ctok 0)) (not (eq? ctok '*eoi*)))
-		      (begin (set! pre-token tok) tok)
-		      (if (null? post-tokens)
-			  (begin (set! pre-token '(*eoi*)) '(*eoi*))
-			  (let ((t (list (car post-tokens)))) 
-			    (set! post-tokens (cdr post-tokens))
-			    (set! pre-token t) 
-			    t)))))))))
-    (lambda ()
-      (catch 'scss:parsed-statement
-	     (lambda () (scss-parser lexer-wrapper parse-error-handler))
-	     (lambda (key . args)
-	       (if error-state
-		   (begin (set! error-state #f) '())
-		   (car args)))))))
  
-(define (parse-fully port)
-  (define (parse-fully-inner stylesheet parser)
-    (let ((st (parser)))
-      (if (not (eq? st '*eoi*))
-	  (begin (if (and st (not (null? st)))
-		     (append! stylesheet `(,st)))
-		 (parse-fully-inner stylesheet parser)))))
-  (let ((s (list 'css))) (parse-fully-inner s (make-parser port)) s))
+(define (parse port)
+  (let* ((lexer (make-css21-lexer port))
+	 (parser (make-css21-parser lexer)))
+    (parser)))
 
 (define (scss:css->scss port . baseurl)
   (define (dirstr str) 
     (let ((s (string-index-right str #\/)))
       (if s (substring str 0 (+ s 1)) "")))
   (define (import-fn s i)
-    (if (not (null? s))
+    (if (and (not (null? s)) (not (null? (cdr s))))
 	(if (eq? (caar s) '@import) (import-fn (cdr s) (cons (cadar s) i)) i)
 	i))
 
@@ -1139,7 +1127,7 @@
   (let* ((bu (if (null? baseurl)
 		 (if (file-port? port) (dirstr (port-filename port)) (getcwd))
 		 (car baseurl)))
-	 (parsed-sheet (parse-fully port))
+	 (parsed-sheet (parse port))
 	 (imports (import-fn (cdr parsed-sheet) (list)))
 	 (clean-sheet (list-tail (cdr parsed-sheet) (length imports)))
 	 (final (fold-right 
@@ -1259,6 +1247,12 @@
 		(background-image background)
 		(background-position background)
 		(background-repeat background)
+
+		(font-family font)
+		(font-size font)
+		(font-style font)
+		(font-variant font)
+		(font-weight font)
 		
 		(margin-top margin)
 		(margin-left margin)
@@ -1346,6 +1340,14 @@
 		       ((equal? ccc "style") (g border-style? cc val))))))
 	  ((equal? c "padding") (g padding? cc val))
 	  ((equal? c "margin") (g margin? cc val))
+
+	  ;; The `background' property is different from some of the other
+	  ;; shorthand properties in that it's possible to give it a value that
+	  ;; doesn't give computed values to all of the more specific
+	  ;; properties for which it is a shorthand.  In these cases, the
+	  ;; specific properties should be explicitly set to their default
+	  ;; values.
+
 	  ((equal? c "background")
 	   (cond ((equal? cc "color") (color? val) val)
 		 ((equal? cc "attach") (member val pattaches))
@@ -1367,12 +1369,28 @@
 		 ((equal? cc "style") (and (border-style? val) val))
 		 ((equal? cc "width") (and (border-width? val) val))
 		 (else #f)))
+
+	  ;; Like `background', the shorthand `font' property sets all values
+	  ;; that can't be deduced from its computed value to their defaults.
+
+	  ;; TODO: Actually parse the computed values!
+
 	  ((equal? c "font")
-	   (cond ((equal? cc "family") (and (font-family? val) val))
-		 ((equal? cc "size") (and (font-size? val) val))
-		 ((equal? cc "style") (and (font-style? val) val))
-		 ((equal? cc "variant") (and (font-variant? val) val))
-		 ((equal? cc "weight") (and (font-weight? val) val))
+	   (cond ((equal? cc "family") 
+		  (or (and (font-family? val) val)
+		      (get-default-prop-value (get-prop-entry 'font-family))))
+		 ((equal? cc "size") 
+		  (or (and (font-size? val) val)
+		      (get-default-prop-value (get-prop-entry 'font-size))))
+		 ((equal? cc "style") 
+		  (or (and (font-style? val) val)
+		      (get-default-prop-value (get-prop-entry 'font-style))))
+		 ((equal? cc "variant") 
+		  (or (and (font-variant? val) val)
+		      (get-default-prop-value (get-prop-entry 'font-variant))))
+		 ((equal? cc "weight") 
+		  (or (and (font-weight? val) val)
+		      (get-default-prop-value (get-prop-entry 'font-weight))))
 		 (else #f)))
 	  (else #f)))))
 
@@ -1380,8 +1398,11 @@
   (define (g x)
     (define (f y)
       (if (eq? (car y) '@media)
-	  (let ((z (cadr y))) 
-	    (and (or (memq medium z) (memq 'all z)) (cddr y)))
+	  (let ((z (cadr y)))
+	    (and (if (list? z)
+		     (or (memq medium z) (memq 'all z))
+		     (or (eq? medium z) (eq? 'all z)))
+		 (cddr y)))
 	  (list y)))
     (let ((r (fold-right append '() (filter-map f (if (null? x) x (cdr x))))))
       (if (null? r) r (cons 'css r))))
@@ -1446,8 +1467,13 @@
       (define (filt y)
 	(define (inner-map r)
 	  (define (inner-map-fn z)
-	    (if (cdr r) (hash-table-set! t (symbol-downcase (car z)) (cdr r)))
-	    (append (cons* sym y z) (list (cdr r))))
+	    (let ((imp (eq? (car z) '!)))
+	      (if (cdr r) 
+		  (hash-table-set! t 
+				   (symbol-downcase (if imp (caadr z) (car z)))
+				   (cdr r)))
+	      (append (if imp (cons* sym y '! (cadr z)) (cons* sym y z))
+		      (list (cdr r)))))
 	  (and r (map inner-map-fn (cdr x))))
 	(inner-map (selector-match? y doc ancestors)))
       (filter-map filt (let ((cx (car x))) (or (grouping? cx #f) (list cx)))))
@@ -1544,6 +1570,8 @@
     (and (pair? x) (pair? y) (eq? (car x) (car y)) (eqv? (cdr x) (cdr y))))
   (define (active-pseudo-elements e)
     (cond ((eq? e doc) 0)
+	  ((internal-pseudo-element-handler "before" doc e) 8)
+	  ((internal-pseudo-element-handler "after" doc e) 4)
 	  ((internal-pseudo-element-handler "first-letter" doc e) 3)
 	  ((internal-pseudo-element-handler "first-line" doc e) 1)
 	  (else (active-pseudo-elements (get-sxml-parent doc e)))))
@@ -1559,7 +1587,7 @@
     (hash-table-ref/default (get-element-hash elt) src #f))
   (define (g x e)
     (and (not (null? x))
-	 (or (hash-table-ref/default e (car x) #f) (g (cdr x)) e)))
+	 (or (hash-table-ref/default e (car x) #f) (g (cdr x) e))))
   (define (f sel)
     (define (cache x h)
       (and x (car (last-pair x)) (hash-table-set! h (car x) x) x))
